@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2009 Canonical Services Ltd (www.canonical.com)
+ * Copyright (C) 2009-2010 Canonical Services Ltd (www.canonical.com)
  *               2009 Mikkel Kamstrup Erlandsen
  *
  * Authors: Rodrigo Moya <rodrigo.moya@canonical.com>
@@ -25,249 +25,145 @@
 #include <libsoup/soup-session-async.h>
 #include <libsoup/soup-gnome.h>
 #include <json-glib/json-glib.h>
+#include "couchdb-database.h"
 #include "couchdb-document.h"
 #include "utils.h"
 
-struct _CouchdbDocument {
-	GObject		parent;
-
-	CouchdbSession  *session;
-	char		*dbname;
-	JsonNode	*root_node;
+struct _CouchdbDocumentPrivate {
+	JsonObject *root_object;
+	CouchdbDatabase *database;
 };
 
 G_DEFINE_TYPE(CouchdbDocument, couchdb_document, G_TYPE_OBJECT);
 
+enum {
+	PROP_0,
+	PROP_DATABASE
+};
+
 static void
 couchdb_document_finalize (GObject *object)
 {
 	CouchdbDocument *document = COUCHDB_DOCUMENT (object);
 
-	g_free (document->dbname);
-	json_node_free (document->root_node);
+	if (document->priv != NULL) {
+		if (document->priv->database != NULL)
+			g_object_unref (G_OBJECT (document->priv->database));
+
+		json_object_unref (document->priv->root_object);
+		g_free (document->priv);
+	}
 
 	G_OBJECT_CLASS (couchdb_document_parent_class)->finalize (object);
 }
 
+static void
+couchdb_document_set_property (GObject *object,
+			       guint prop_id,
+			       const GValue *value,
+			       GParamSpec *pspec)
+{
+	CouchdbDocument *document = COUCHDB_DOCUMENT (object);
+
+	switch (prop_id) {
+	case PROP_DATABASE:
+		if (document->priv->database != NULL)
+			g_object_unref (G_OBJECT (document->priv->database));
+		document->priv->database = g_object_ref (G_OBJECT (g_value_get_object (value)));
+		break;
+	default:
+                G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+                break;
+	}
+}
+
+static void
+couchdb_document_get_property (GObject *object,
+                               guint prop_id,
+                               GValue *value,
+                               GParamSpec *pspec)
+{
+	CouchdbDocument *document = COUCHDB_DOCUMENT (object);
+
+	switch (prop_id) {
+	case PROP_DATABASE:
+		g_value_set_object (value, document->priv->database);
+		break;
+	default:
+                G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+                break;
+	}
+}
+
 static void
 couchdb_document_class_init (CouchdbDocumentClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
 	object_class->finalize = couchdb_document_finalize;
+	object_class->set_property = couchdb_document_set_property;
+	object_class->get_property = couchdb_document_get_property;
+
+	g_object_class_install_property (object_class,
+					 PROP_DATABASE,
+					 g_param_spec_object ("database",
+							      "Database object",
+							      "Database from which this document was retrieved/saved",
+							      COUCHDB_TYPE_DATABASE,
+							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
 }
 
 static void
 couchdb_document_init (CouchdbDocument *document)
 {
-	document->session = NULL;
-	document->root_node = NULL;
-	document->dbname = NULL;
+	document->priv = g_new0 (CouchdbDocumentPrivate, 1);
+	document->priv->root_object = json_object_new ();
 }
 
 /**
  * couchdb_document_new:
- * @couchdb: A #CouchdbSession object
  *
  * Create an empty #CouchdbDocument object, which can then be populated with data
- * and added to a database on the specified CouchDB instance.
+ * and added to a database.
  *
  * Return value: A newly-created #CouchdbDocument object, with no data on it.
  */
 CouchdbDocument *
-couchdb_document_new (CouchdbSession *session)
+couchdb_document_new (void)
 {
 	CouchdbDocument *document;
 
 	document = g_object_new (COUCHDB_TYPE_DOCUMENT, NULL);
-	document->session = session;
-	document->dbname = NULL;
-
-	document->root_node = json_node_new (JSON_NODE_OBJECT);
-	json_node_set_object (document->root_node, json_object_new ());
 
 	return document;
 }
 
 CouchdbDocument *
-couchdb_document_new_from_json_object (CouchdbSession *session, JsonObject *json_object)
+couchdb_document_new_from_json_object (JsonObject *json_object)
 {
 	CouchdbDocument *document;
 
 	document = g_object_new (COUCHDB_TYPE_DOCUMENT, NULL);
-	document->session = session;
-	document->dbname = NULL;
 
-	document->root_node = json_node_new (JSON_NODE_OBJECT);
-	json_node_set_object (document->root_node, json_object_ref (json_object));
+	if (document->priv->root_object != NULL)
+		json_object_unref (document->priv->root_object);
+
+	document->priv->root_object = json_object_ref (json_object);
 
 	return document;
 }
 
-/**
- * couchdb_document_get:
- * @couchdb: A #CouchdbSession object
- * @dbname: Name of the database to retrieve the document from
- * @docid: Unique ID of the document to be retrieved
- * @error: Placeholder for error information
- *
- * Retrieve the last revision of a document from the given database.
- *
- * Return value: A #CouchdbDocument object if successful, NULL otherwise, in
- * which case, the error argument will contain information about the error.
- */
-CouchdbDocument *
-couchdb_document_get (CouchdbSession *session,
-		      const char *dbname,
-		      const char *docid,
-		      GError **error)
+void
+couchdb_document_set_from_json_object (CouchdbDocument *document, JsonObject *json_object)
 {
-	char *url, *encoded_docid;
-	JsonParser *parser;
-	CouchdbDocument *document = NULL;
+	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
 
-	g_return_val_if_fail (COUCHDB_IS_SESSION (session), NULL);
-	g_return_val_if_fail (dbname != NULL, NULL);
-	g_return_val_if_fail (docid != NULL, NULL);
+	if (document->priv->root_object != NULL)
+		json_object_unref (document->priv->root_object);
 
-	encoded_docid = soup_uri_encode (docid, NULL);
-	url = g_strdup_printf ("%s/%s/%s", couchdb_session_get_uri (session), dbname, encoded_docid);
-	parser = json_parser_new ();
-	if (couchdb_session_send_message (session, SOUP_METHOD_GET, url, NULL, parser, error)) {
-		document = g_object_new (COUCHDB_TYPE_DOCUMENT, NULL);
-		document->session = session;
-		document->dbname = g_strdup (dbname);
-
-		document->root_node = json_node_copy (json_parser_get_root (parser));		
-	}
-	g_object_unref (G_OBJECT (parser));
-	g_free (encoded_docid);
-	g_free (url);
-
-	return document;
+	document->priv->root_object = json_object_ref (json_object);
 }
 
-/**
- * couchdb_document_put:
- * @document: A #CouchdbDocument object
- * @dbname: Name of the database where the document will be stored
- * @error: Placeholder for error information
- *
- * Store a document on a CouchDB database.
- *
- * If it is a new document, and hence does not have a unique ID, a unique ID
- * will be generated and stored on the #CouchdbDocument object. Likewise,
- * whether the document is new or just an update to an existing one, the
- * #CouchdbDocument object passed to this function will be updated to contain
- * the latest revision of the document, as returned by CouchDB (revision that
- * can be retrieved by calling #couchdb_document_get_revision).
- *
- * Return value: TRUE if successful, FALSE otherwise, in which case the error
- * argument will contain information about the error.
- */
-gboolean
-couchdb_document_put (CouchdbDocument *document,
-		      const char *dbname,
-		      GError **error)
-{
-	char *url, *body;
-	const char *id;
-	JsonParser *parser;
-	gboolean result = FALSE;
-	gboolean send_ok;
-
-	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), FALSE);
-	g_return_val_if_fail (dbname != NULL, FALSE);
-
-	id = couchdb_document_get_id (document);
-	body = couchdb_document_to_string (document);
-	parser = json_parser_new ();
-	if (id) {
-		char *encoded_docid;
-
-		encoded_docid = soup_uri_encode (id, NULL);
-		url = g_strdup_printf ("%s/%s/%s", couchdb_session_get_uri (document->session), dbname, encoded_docid);		
-		send_ok = couchdb_session_send_message (document->session, SOUP_METHOD_PUT, url, body, parser, error);
-
-		g_free (encoded_docid);
-	} else {
-		url = g_strdup_printf ("%s/%s/", couchdb_session_get_uri (document->session), dbname);
-		send_ok = couchdb_session_send_message (document->session, SOUP_METHOD_POST, url, body, parser, error);
-	}
-
-	if (send_ok) {
-		JsonObject *object;
-
-		object = json_node_get_object (json_parser_get_root (parser));
-		couchdb_document_set_id (document, json_object_get_string_member (object, "id"));
-		couchdb_document_set_revision (document, json_object_get_string_member (object, "rev"));
-
-		if (document->dbname) {
-			g_free (document->dbname);
-			document->dbname = g_strdup (dbname);
-		}
-
-		if (id)
-			g_signal_emit_by_name (document->session, "document_updated", dbname, document);
-		else
-			g_signal_emit_by_name (document->session, "document_created", dbname, document);
-
-		result = TRUE;
-	}
-
-	/* free memory */
-	g_free (url);
-	g_free (body);
-	g_object_unref (G_OBJECT (parser));
-
-	return result;
-}
-
-/**
- * couchdb_document_delete:
- * @document: A #CouchdbDocument object
- * @error: Placeholder for error information
- *
- * Delete an existing document from a CouchDB instance.
- *
- * Please take note that this operation can fail if there was an update to the
- * document and that last revision was not retrieved by the calling application.
- * This is due to the fact that, to remove a document from CouchDB, the latest
- * revision needs to be sent, so if the #CouchdbDocument object passed to this
- * function does not contain the last revision, the operation will fail. In that
- * case, retrieving the latest revision from CouchDB (with #couchdb_document_get)
- * and trying the delete operation again should fix the issue.
- *
- * Return value: TRUE if successful, FALSE otherwise, in which case the error
- * argument will contain information about the error.
- */
-gboolean
-couchdb_document_delete (CouchdbDocument *document, GError **error)
-{
-	const char *id, *revision;
-	char *url;
-	JsonParser *parser;
-	gboolean result = FALSE;
-
-	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), FALSE);
-
-	id = couchdb_document_get_id (document);
-	revision = couchdb_document_get_revision (document);
-	if (!id || !revision) /* we can't remove a document without an ID and/or a REVISION */
-		return FALSE;
-
-	url = g_strdup_printf ("%s/%s/%s?rev=%s", couchdb_session_get_uri (document->session), document->dbname, id, revision);
-	
-	/* We don't parse the http response, therefore the parser arg is NULL */
-	if (couchdb_session_send_message (document->session, SOUP_METHOD_DELETE, url, NULL, NULL, error)) {
-		result = TRUE;		
-		g_signal_emit_by_name (document->session, "document_deleted", document->dbname, id);
-	}
-
-	g_free (url);
-
-	return result;
-}
 
 /**
  * couchdb_document_get_id:
@@ -282,12 +178,10 @@ couchdb_document_get_id (CouchdbDocument *document)
 {
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
 
-	if (document->root_node &&
-	    json_node_get_node_type (document->root_node) == JSON_NODE_OBJECT) {
-		if (json_object_has_member (json_node_get_object (document->root_node),
-					    "_id"))
+	if (document->priv->root_object != NULL) {
+		if (json_object_has_member (document->priv->root_object, "_id"))
 			return json_object_get_string_member (
-				json_node_get_object (document->root_node),
+				document->priv->root_object,
 				"_id");
 	}
 
@@ -313,9 +207,7 @@ couchdb_document_set_id (CouchdbDocument *document, const char *id)
 	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
 	g_return_if_fail (id != NULL);
 
-	json_object_set_string_member (json_node_get_object (document->root_node),
-				       "_id",
-				       id);
+	json_object_set_string_member (document->priv->root_object, "_id", id);
 }
 
 /**
@@ -337,10 +229,9 @@ couchdb_document_get_revision (CouchdbDocument *document)
 {
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
 
-	if (document->root_node &&
-	    json_node_get_node_type (document->root_node) == JSON_NODE_OBJECT) {
+	if (document->priv->root_object != NULL) {
 		return json_object_get_string_member (
-			json_node_get_object (document->root_node),
+			document->priv->root_object,
 			"_rev");
 	}
 
@@ -362,9 +253,7 @@ couchdb_document_set_revision (CouchdbDocument *document, const char *revision)
 	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
 	g_return_if_fail (revision != NULL);
 
-	json_object_set_string_member (json_node_get_object (document->root_node),
-				       "_rev",
-				       revision);
+	json_object_set_string_member (document->priv->root_object, "_rev", revision);
 }
 
 /**
@@ -379,12 +268,10 @@ couchdb_document_set_revision (CouchdbDocument *document, const char *revision)
 gboolean
 couchdb_document_has_field (CouchdbDocument *document, const char *field)
 {
-	JsonNode *node;
-
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), FALSE);
 	g_return_val_if_fail (field != NULL, FALSE);
 
-	return json_object_has_member (json_node_get_object (document->root_node), field);
+	return json_object_has_member (document->priv->root_object, field);
 }
 
 /**
@@ -400,7 +287,52 @@ couchdb_document_remove_field (CouchdbDocument *document, const char *field)
 	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
 	g_return_if_fail (field != NULL);
 
-	json_object_remove_member (json_node_get_object (document->root_node), field);
+	json_object_remove_member (document->priv->root_object, field);
+}
+
+/**
+ * couchdb_document_get_field_names:
+ * @document: A #CouchdbDocument object
+ *
+ * Retrieve the list of fields contained in the given #CouchdbDocument.
+ *
+ * Return value: A list of strings containing the names of all the fields contained
+ * in the given #CouchdbDocument object. When no longer needed, the list should
+ * be freed by calling #g_slist_free.
+ */
+GSList *
+couchdb_document_get_fields (CouchdbDocument *document)
+{
+	GList *json_list;
+	GSList *result = NULL;
+
+	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
+
+	json_list = json_object_get_members (document->priv->root_object);
+	while (json_list != NULL) {
+		result = g_slist_append (result, json_list->data);
+
+		json_list = g_list_remove (json_list, json_list->data);
+	}
+
+	return result;
+}
+
+/**
+ * couchdb_document_get_field_type:
+ * @document: A #CouchdbDocument object
+ * @field: Name of the field to get type
+ *
+ * Get the value type of the given #CouchdbDocument's field.
+ *
+ * Return value: Type of the field.
+ */
+GType
+couchdb_document_get_field_type (CouchdbDocument *document, const char *field)
+{
+	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), G_TYPE_INVALID);
+
+	return get_type_from_field (document->priv->root_object, field);
 }
 
 /**
@@ -418,12 +350,12 @@ couchdb_document_get_array_field (CouchdbDocument *document, const char *field)
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
 	g_return_val_if_fail (field != NULL, NULL);
 
-	if (!json_object_has_member (json_node_get_object (document->root_node), field))
+	if (!json_object_has_member (document->priv->root_object, field))
 		return NULL;
 
 	return couchdb_array_field_new_from_json_array (
-		json_array_ref (json_object_get_array_member (json_node_get_object (document->root_node),
-							       field)));
+		json_array_ref (json_object_get_array_member (document->priv->root_object,
+							      field)));
 }
 
 /**
@@ -441,7 +373,7 @@ couchdb_document_set_array_field (CouchdbDocument *document, const char *field,
 	g_return_if_fail (field != NULL);
 	g_return_if_fail (value != NULL);
 
-	json_object_set_array_member (json_node_get_object (document->root_node),
+	json_object_set_array_member (document->priv->root_object,
 				      field,
 				      json_array_ref (couchdb_array_field_get_json_array (value)));
 }
@@ -461,10 +393,10 @@ couchdb_document_get_boolean_field (CouchdbDocument *document, const char *field
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), FALSE);
 	g_return_val_if_fail (field != NULL, FALSE);
 
-	if (!json_object_has_member (json_node_get_object (document->root_node), field))
+	if (!json_object_has_member (document->priv->root_object, field))
 		return FALSE;
 
-	return json_object_get_boolean_member (json_node_get_object (document->root_node),
+	return json_object_get_boolean_member (document->priv->root_object,
 					       field);
 }
 
@@ -482,9 +414,7 @@ couchdb_document_set_boolean_field (CouchdbDocument *document, const char *field
 	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
 	g_return_if_fail (field != NULL);
 
-	json_object_set_boolean_member (json_node_get_object (document->root_node),
-					field,
-					value);
+	json_object_set_boolean_member (document->priv->root_object, field, value);
 }
 
 /**
@@ -502,10 +432,10 @@ couchdb_document_get_int_field (CouchdbDocument *document, const char *field)
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), -1);
 	g_return_val_if_fail (field != NULL, -1);
 
-	if (!json_object_has_member (json_node_get_object (document->root_node), field))
+	if (!json_object_has_member (document->priv->root_object, field))
 		return 0;
 
-	return json_object_get_int_member (json_node_get_object (document->root_node),
+	return json_object_get_int_member (document->priv->root_object,
 					   field);
 }
 
@@ -523,9 +453,7 @@ couchdb_document_set_int_field (CouchdbDocument *document, const char *field, gi
 	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
 	g_return_if_fail (field != NULL);
 
-	json_object_set_int_member (json_node_get_object (document->root_node),
-				    field,
-				    value);
+	json_object_set_int_member (document->priv->root_object, field, value);
 }
 
 /**
@@ -543,10 +471,10 @@ couchdb_document_get_double_field (CouchdbDocument *document, const char *field)
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), -1);
 	g_return_val_if_fail (field != NULL, -1);
 
-	if (!json_object_has_member (json_node_get_object (document->root_node), field))
+	if (!json_object_has_member (document->priv->root_object, field))
 		return 0.0;
-	return json_object_get_double_member (json_node_get_object (document->root_node),
-					      field);
+
+	return json_object_get_double_member (document->priv->root_object, field);
 }
 
 /**
@@ -563,9 +491,7 @@ couchdb_document_set_double_field (CouchdbDocument *document, const char *field,
 	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
 	g_return_if_fail (field != NULL);
 
-	json_object_set_double_member (json_node_get_object (document->root_node),
-				       field,
-				       value);
+	json_object_set_double_member (document->priv->root_object, field, value);
 }
 
 /**
@@ -583,11 +509,10 @@ couchdb_document_get_string_field (CouchdbDocument *document, const char *field)
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
 	g_return_val_if_fail (field != NULL, NULL);
 
-	if (!json_object_has_member (json_node_get_object (document->root_node), field))
+	if (!json_object_has_member (document->priv->root_object, field))
 		return NULL;
 
-	return json_object_get_string_member (json_node_get_object (document->root_node),
-					      field);
+	return json_object_get_string_member (document->priv->root_object, field);
 }
 
 /**
@@ -605,7 +530,7 @@ couchdb_document_set_string_field (CouchdbDocument *document, const char *field,
 	g_return_if_fail (field != NULL);
 
 	if (value) {
-		json_object_set_string_member (json_node_get_object (document->root_node),
+		json_object_set_string_member (document->priv->root_object,
 					       field,
 					       value);
 	} else {
@@ -629,11 +554,11 @@ couchdb_document_get_struct_field (CouchdbDocument *document, const char *field)
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
 	g_return_val_if_fail (field != NULL, NULL);
 
-	if (!json_object_has_member (json_node_get_object (document->root_node), field))
+	if (!json_object_has_member (document->priv->root_object, field))
 		return NULL;
 
 	return couchdb_struct_field_new_from_json_object (
-		json_object_ref (json_object_get_object_member (json_node_get_object (document->root_node),
+		json_object_ref (json_object_get_object_member (document->priv->root_object,
 								field)));
 }
 
@@ -652,11 +577,43 @@ couchdb_document_set_struct_field (CouchdbDocument *document, const char *field,
 	g_return_if_fail (field != NULL);
 	g_return_if_fail (value != NULL);
 
-	json_object_set_object_member (json_node_get_object (document->root_node),
+	json_object_set_object_member (document->priv->root_object,
 				       field,
 				       json_object_ref (couchdb_struct_field_get_json_object (value)));
 }
 
+/**
+ * couchdb_document_foreach_attachment:
+ * @document: A #CouchdbDocument object
+ * @func: Callback function to be called for each attachment found
+ * @user_data: Data to pass to callback function
+ *
+ * List all the attachments for the given document.
+ */
+void
+couchdb_document_foreach_attachment (CouchdbDocument *document,
+				     CouchdbDocumentForeachAttachmentFunc func,
+				     gpointer user_data)
+{
+	CouchdbStructField *attachments;
+
+	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
+
+	attachments = couchdb_document_get_struct_field (document, "_attachments");
+	if (attachments != NULL) {
+		GSList *list;
+
+		list = couchdb_struct_field_get_field_names (attachments);
+		while (list != NULL) {
+			func (document, (const gchar *) list->data, user_data);
+
+			list = g_slist_remove (list, list->data);
+		}
+
+		g_object_unref (G_OBJECT (attachments));
+	}
+}
+
 /**
  * couchdb_document_to_string:
  * @document: A #CouchdbDocument object
@@ -670,16 +627,20 @@ couchdb_document_to_string (CouchdbDocument *document)
 {
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
 
-	if (document->root_node) {
+	if (document->priv->root_object) {
 		JsonGenerator *generator;
 		char *str;
 		gsize size;
+		JsonNode *root_node;
 
 		generator = json_generator_new ();
-		json_generator_set_root (generator, document->root_node);
+		root_node = json_node_new (JSON_NODE_OBJECT);
+		json_node_set_object (root_node, json_object_ref (document->priv->root_object));
+		json_generator_set_root (generator, root_node);
 
 		str = json_generator_to_data (generator, &size);
 		g_object_unref (G_OBJECT (generator));
+		json_node_free (root_node);
 
 		return str;
 	}
@@ -692,5 +653,113 @@ couchdb_document_get_json_object (CouchdbDocument *document)
 {
 	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
 	
-	return json_node_get_object (document->root_node);
+	return document->priv->root_object;
+}
+
+/**
+ * couchdb_document_get_record_type:
+ * @document: A #CouchdbDocument object
+ *
+ * Retrieve the record type of the given document. Record types are special
+ * fields in the CouchDB documents, used in Desktopcouch, to identify
+ * standard records. All supported record types are listed at
+ * http://www.freedesktop.org/wiki/Specifications/desktopcouch#Formats.
+ *
+ * Return value: The record type of the given document.
+ */
+const char *
+couchdb_document_get_record_type (CouchdbDocument *document)
+{
+	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
+
+	return couchdb_document_get_string_field (COUCHDB_DOCUMENT (document), "record_type");
+}
+
+/**
+ * couchdb_document_set_record_type:
+ * @document: A #CouchdbDocument object
+ * @record_type: Record type to set the document to
+ *
+ * Set the record type of the given document.
+ */
+void
+couchdb_document_set_record_type (CouchdbDocument *document, const char *record_type)
+{
+	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
+	g_return_if_fail (record_type != NULL);
+
+	couchdb_document_set_string_field (COUCHDB_DOCUMENT (document), "record_type", record_type);
+}
+
+/**
+ * couchdb_document_get_application_annotations:
+ * @document: A #CouchdbDocument object
+ *
+ * Retrieve the application annotations for the given document.
+ *
+ * Application annotations is a special field (named "application_annotations"), used
+ * in Desktopcouch to allow applications to set values on standard documents (as defined
+ * at http://www.freedesktop.org/wiki/Specifications/desktopcouch#Formats) that are
+ * not part of the standard, but still needed by the application.
+ *
+ * Return value: A #CouchdbStructField containing the value of the application
+ * annotations for the given document.
+ */
+CouchdbStructField *
+couchdb_document_get_application_annotations (CouchdbDocument *document)
+{
+	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), NULL);
+
+	return couchdb_document_get_struct_field (COUCHDB_DOCUMENT (document), "application_annotations");
+}
+
+/**
+ * couchdb_document_set_application_annotations:
+ * @document: A #CouchdbDocument object
+ * @annotations: A #CouchdbStructField with the contents of the application_annotations field.
+ *
+ * Set the application annotations for the given document.
+ */
+void
+couchdb_document_set_application_annotations (CouchdbDocument *document, CouchdbStructField *annotations)
+{
+	g_return_if_fail (COUCHDB_IS_DOCUMENT (document));
+
+	couchdb_document_set_struct_field (COUCHDB_DOCUMENT (document), "application_annotations", annotations);
+}
+
+/**
+ * couchdb_document_is_contact:
+ * @document: A #CouchdbDocument object
+ *
+ * Check whether the given document represents a contact record, as specified
+ * at http://www.freedesktop.org/wiki/Specifications/desktopcouch/contact
+ *
+ * Return value: TRUE if the document represents a contact, FALSE otherwise.
+ */
+gboolean
+couchdb_document_is_contact (CouchdbDocument *document)
+{
+	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), FALSE);
+
+	return !g_ascii_strcasecmp (couchdb_document_get_record_type (document),
+				    COUCHDB_RECORD_TYPE_CONTACT);
+}
+
+/**
+ * couchdb_document_is_task:
+ * @document: A #CouchdbDocument object
+ *
+ * Check whether the given document represents a contact record, as specified
+ * at http://www.freedesktop.org/wiki/Specifications/desktopcouch/task
+ *
+ * Return value: TRUE if the document represents a task, FALSE otherwise.
+ */
+gboolean
+couchdb_document_is_task (CouchdbDocument *document)
+{
+	g_return_val_if_fail (COUCHDB_IS_DOCUMENT (document), FALSE);
+
+	return !g_ascii_strcasecmp (couchdb_document_get_record_type (document),
+				    COUCHDB_RECORD_TYPE_TASK);
 }
