@@ -28,8 +28,12 @@
 #include <couchdb-array-field.h>
 #include <couchdb-credentials.h>
 #include <couchdb-database-info.h>
+#include <couchdb-database.h>
+#include <couchdb-design-document.h>
 #include <couchdb-document.h>
+#include <couchdb-document-contact.h>
 #include <couchdb-document-info.h>
+#include <couchdb-document-task.h>
 #include <couchdb-session.h>
 #include <couchdb-struct-field.h>
 
